// DONE find the 'TODOs' below and do them!

#include <assert.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>

typedef struct Gate Gate;

struct Gate
{
  int var;	// Tseitin variable index of this gate
  int input;	// non-zero if this is an input gate
  int used;	// non-zero if this gate has been used
};

int n;		// number of inputs
int m;		// number of gates
int v;		// number of Tseitin variables

Gate * gates;

// Print binary or ternary DIMACS clauses to 'stdout'.
// DONE use these where clauses are supposed to be printed.

void binary (int a, int b) { printf ("%d %d 0\n", a, b); }
void ternary (int a, int b, int c) { printf ("%d %d %d 0\n", a, b, c); }

// Check whether 'lit' is a valid Tseitin literal.

int valid (int lit) { return lit && lit != INT_MIN && abs (lit) <= v; }

// Constructor for a new input gate.

void input ()
{
  Gate * g = gates + m++;
  assert (!g->used);
  g->input = 1;
  g->var = m;
}

// Constructor for a new XOR gate.

void xor (Gate * first, Gate * second)
{
  // Every input and XOR gate should be used at most once, actually exactly
  // once except for the final output of the parity circuit.

  assert (!first->used);
  assert (!second->used);

  assert (first != second);	// Require different inputs to this gate.

  Gate * g = gates + m++;
  assert (!g->input);
  assert (!g->used);
  g->var = ++v;

  // DONE print clauses for this XOR gate given the Tseitin variables
  //
  //   g->var, first->var, second->var
  //
  // so in essence a CNF for the following constraint
  //
  //   'g->var' == 'first->var' ^ 'second->var'

  ternary (-g->var, -first->var, second->var);
  ternary (-g->var, first->var, -second->var);
  ternary (g->var, -first->var, -second->var);
  ternary (g->var, first->var, second->var);

  first->used = second->used = 1;	// mark 'first' and 'second' as used
}

// Generate a random parity checking circuit and print its Tseitin encoding.
// Returns the Tseitin literal for the output of the parity circuit.

int parity ()
{
  // Reset and allocate enough gates to hold a parity tree ('2*n - 1').

  m = 0;
  gates = calloc (2*n - 1, sizeof *gates);

  // First encode 'n' input gates.

  int i;
  for (i = 1; i <= n; i++) input (i);

  // Now encode  'n-1' XOR gates made of unused gates as input.

  for (i = 1; i < n; i++) {
    int first, second;
    // DONE choose 'first' and 'second' randomly to make the circuits different.
    // Hint: pick two random existing unused gates from gates[0]...gates[v-1].
    do first = rand () % m; while (gates[first].used);
    do second = rand () % m; while (first == second || gates[second].used);
    xor (gates + first, gates + second);
  }
  free (gates);

  return v;	// Return Tseitin variable of last added gate.
}

void die (char * m) { fprintf (stderr, "error: %s\n", m); exit (1); }

int main (int argc, char ** argv)
{
  // Option parsing.

  if (argc < 2 || argc > 3) die ("usage: <inputs> [ <seed> ]");
  n = atoi (argv[1]);
  if (n < 2) die ("require at least two inputs");
  printf ("c inputs %d\n", n);
  int seed = argc > 3 ? atoi (argv[2]) : 0;
  printf ("c seed %d\n", seed);
  srand (seed);

  // Initialize Tseitin encoding.

  v = n;		// Inputs get first 'n' Tseitin variables.

  // Print DIMACS header.

  // DONE How many variables and clauses are needed (instead of '0')?

  printf ("p cnf %d %d\n", 3*n-2, 8*(n-1) + 2);

  // Construct both parity circuits and print their Tseitin encoding.

  int first = parity ();
  int second = parity ();

  // DONE print clauses to force  that 'first' is different from 'second'.
  // Thus in essence encode the XOR between the Tseitin variables
  //
  //   first, second
  //
  // or as constraint encode
  //
  //   'first' ^ 'second'

  binary (-first, -second);
  binary (first, second);

  return 0;
}
